import roifile
import numpy as np
from math import sqrt, pow, floor
from numpy import array, zeros

"""
Set of functions to handle ROIs files in DeXtrusion (compare, save and write)


BSD 3-Clause License

Copyright (c) 2022, Gaëlle  LETORT and Alexis Villars
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.
"""

def write_rois(outfile, rois, verbose=True):
    if verbose:
        print("Writing "+str(len(rois))+" Rois in file "+outfile)
    roifile.roiwrite(outfile, rois, mode='w')

def create_roi(pt, cat=1, astime=True):
    """ Create an ImageJ Roi from point coordinates """
    croi = roifile.ImagejRoi()
    croi.version = 227
    croi.roitype = roifile.ROI_TYPE(10)
    #croi.name = str(val)
    croi.name = str(pt[0]+1).zfill(4)+'-'+str(pt[1]).zfill(4)+"-"+str(pt[2]).zfill(4)
    croi.n_coordinates = 1
    croi.left = int(pt[2])
    croi.top = int(pt[1])
    if astime:
        croi.z_position = 1
        croi.t_position = pt[0]+1
    else:
        croi.z_position = pt[0]+1
        croi.t_position = 1
    croi.c_position = 1
    croi.integer_coordinates = array([0,0])
    croi.stroke_width=3
    if cat == 1:  ## color cell extrusion
        croi.stroke_color = b'\xff\xff\x00\x00'
    if cat == 2:  ## color cell sop
        croi.stroke_color = b'\xff\x00\x00\xff'
    if cat == 3:  ## color cell division
        croi.stroke_color = b'\xff\x00\xff\x00'
    return croi

def distance_rois(roi1, roi2):
    return sqrt( pow(roi1[2]-roi2[2],2) + pow(roi1[1]-roi2[1],2) )
    
def matched_roi(roi, lroi, dxy, dt, matched):
    closest = -1
    mind = dt+dxy*1.1
    for i in range(len(lroi)):
        curroi = lroi[i]
        if abs(curroi[0]-roi[0]) <= dt:
            if distance_rois(curroi, roi) <= dxy:
                if matched is None:
                    return True
                if (matched[i]==0):
                    mixdist = distance_rois(curroi, roi)+abs(curroi[0]-roi[0])
                    if mixdist < mind:
                        mind = mixdist
                        closest = i
    if closest < 0:
        return False

    if matched is not None:
        matched[closest] = 1
        return True
    return True
    
def read_rois(resfile):
    ''' read an imageJ Rois file and return the list (z, y, x) of the rois coordinates '''
    rois =  roifile.ImagejRoi.fromfile(resfile)
    listrois = []
    for roi in rois:
        xroi = roi.left
        yroi = roi.top
        frame = max(roi.position, roi.z_position, roi.t_position)-1
        #print(str(roi.position)+" "+str(roi.t_position)+" "+str(roi.z_position)+" "+str(frame))
        listrois.append((frame, yroi, xroi))
    return listrois

def clean_rois(resfile, volthres=10, probthres=180, writefile=True, verbose=True):
    ''' read an imageJ results Rois file and return the list (z, y, x) of the rois coordinates that are above given thresholds '''
    rois =  roifile.ImagejRoi.fromfile(resfile)
    listrois = []
    for roi in rois:
        xroi = roi.left
        yroi = roi.top
        frame = max(roi.z_position, roi.t_position)-1
        name = roi.name
        #print(str(roi.position)+" "+str(roi.t_position)+" "+str(roi.z_position)+" "+str(frame))
        namepart = name.split("-")
        if (float(namepart[3]) > volthres) and (float(namepart[4]) > probthres):
            if writefile:
                listrois.append(roi)
            else:
                listrois.append((frame, yroi, xroi))
    if writefile:
        write_rois(resfile+"_cleaned.zip", listrois, verbose)
        return None
    else:
        return listrois


def get_falsepositives(mrois, grois, distance_xy, distance_t):
    fp = []
    matched = zeros(len(grois))
    for roi in mrois:
        if not matched_roi(roi, grois, distance_xy, distance_t, matched):
            fp.append(create_roi((roi[0], roi[1], roi[2])))
    return fp
    
def get_falsenegatives(mrois, grois, distance_xy, distance_t):
    fn = []
    for roi in grois:
        if not matched_roi(roi, mrois, distance_xy, distance_t, None):
            fn.append(create_roi((roi[0], roi[1], roi[2])))
    return fn

def check_positives(mrois, grois, distance_xy, distance_t):
    tp = 0
    fp = 0
    matched = zeros(len(grois))
    for roi in mrois:
        if matched_roi(roi, grois, distance_xy, distance_t, matched):
            tp += 1
        else:
            fp += 1
    return (tp, fp)

def check_falseneg(mrois, grois, distance_xy, distance_t):
    fp = 0
    for roi in mrois:
        if not matched_roi(roi, grois, distance_xy, distance_t, None):
            fp += 1
    return fp
    
def write_falsenegatives(resfile, gtfile, outputfile, distance_xy=10, distance_t=3):
        ''' Compare Rois position: count number of FP, FN '''
        myrois = read_rois(resfile)
        gtrois = read_rois(gtfile)
        
        false_neg = get_falsenegatives(myrois, gtrois, distance_xy, distance_t)
        roifile.roiwrite(outputfile, false_neg, mode='w')
    
def write_falsepositives(resfile, gtfile, outputfile, distance_xy=10, distance_t=3):
        ''' Compare Rois position: count number of FP, FN '''
        myrois = read_rois(resfile)
        gtrois = read_rois(gtfile)
        
        false_pos = get_falsepositives(myrois, gtrois, distance_xy, distance_t)
        roifile.roiwrite(outputfile, false_pos, mode='w')

def compare_rois(resfile, gtfile, distance_xy=10, distance_t=3):
        ''' Compare Rois position: count number of FP, FN '''
        myrois = read_rois(resfile)
        gtrois = read_rois(gtfile)
        
        true_pos, false_pos = check_positives(myrois, gtrois, distance_xy, distance_t)
        false_neg = check_falseneg(gtrois, myrois, distance_xy, distance_t)
        print("True positives "+str(true_pos))
        print("False positives "+str(false_pos))
        print("False negatives "+str(false_neg))
        if (true_pos+false_pos) == 0:
            print("Precision "+str(0))
            print("Recall "+str(0))
            return [true_pos, false_pos, false_neg, 0, 0]
        print("Precision "+str(true_pos/(true_pos+false_pos)))
        print("Recall "+str(true_pos/(true_pos+false_neg)))
        return [true_pos, false_pos, false_neg, true_pos/(true_pos+false_pos), true_pos/(true_pos+false_neg)]
    



import functools
import os, random, re, time, cv2, gc
import numpy as np
from math import floor, ceil, isnan
from numpy import zeros, uint8, expand_dims
from numpy import min as npmin
from numpy import max as npmax
from tifffile import imread, imwrite
from skimage.segmentation import watershed
from skimage.morphology import local_maxima, disk
from skimage.filters import rank
from scipy.ndimage import zoom, label, distance_transform_edt, maximum_filter, minimum_filter, median_filter, uniform_filter
from scipy.ndimage import measurements, zoom
from scipy.ndimage import sum as ndsum

import dextrusion.RoiUtils as ru
from dextrusion.Network import Network  

"""
BSD 3-Clause License

Copyright (c) 2022, Gaëlle  LETORT and Alexis Villars
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

class DeXtrusion:
    """

        Detect extrusion/other events position in a whole movie 
        
        - (Re)Train a neural network to do the detection
        - Use trained neural network(s) on a movie to detect events

        DeXtrusion uses moving windows and used the trained network(s) on each window to calculate its probability to contain an event.
        The probability map can be converted to a list of point events, and exported as Fiji ROIs.

        DeXtrusion is distributed as a python module, installable via pip or via the setup file.

    """

    def __init__(self, verbose=True):
        """ Create class object. 
        :param verbose: if should print information messages while running
        """
        self.model = Network(verbose)    ## initialize neural network     
        self.verbose = verbose           ## print information messages while running   
        self.clear()

    def clear(self):
        """ Reset the object """
        self.nframes = (0,0)
        self.half_size = (0,0)
        self.ncat = 0
        self.catnames = None
        self.nfilters = 0
        self.batch_size = 0
        self.cell_diameter = 0
        self.extrusion_duration = 0
        

    ################################################################################"
    ###### Parameters/configuration

    def set_parameters(self, nframes=(5,5), half_size=(22,22), ncat=4, catnames=None, nb_filters=8, batch_size=30, model_path="./model", cell_diameter=25, extrusion_duration=4.5):
        """ Set the parameters of deXtrusion network, DeXNet
        :param nframes: number of frames before and after the event to use in the sliding windows
        :param half_size: half size of the sliding window in X and Y
        :param ncat: number of event type to detect
        :param catnames: the names of the events (should be the same as the end of the training ROI files)
        :param nb_filters: size of the convolutional neural network before going through the temporal network
        :param model_path: path to where the model is saved/to be saved
        :param batch_size: batch size used for training of the network
        :param cell_diameter: typical diameter of a cell, in pixels, used to resize the movies if necessary
        :param extrusion_duration: typical number of frames in which the extrusion is visible, used to resize temporally the movies f necessary
        """
        self.nframes = nframes
        self.half_size = half_size
        self.ncat = ncat
        self.catnames = catnames
        if self.catnames is None: 
            self.catnames = ["", "_cell_death.zip", "_cell_sop.zip", "_cell_division.zip"]
        self.nfilters = nb_filters
        self.batch_size = batch_size
        self.model_path = model_path
        self.cell_diameter = cell_diameter
        self.extrusion_duration = extrusion_duration
        
        self.write_configuration()  ## write a file with all these paramters to allow to reload it
    
    def retrain_parameters( self, model_path="./model_retrain", datapath="./", aug=3, epoch=10 ):
        """ Load network parameters and change only those to retrain
        """
        self.model_path = model_path
        self.write_configuration()
        self.add_configuration_info( epoch, aug, path=datapath )
        
    def write_configuration(self):
        """ Write the values of the parameter used in a configuration file (named config.cfg) to allow to load the network later """
        if not os.path.exists(self.model_path):
            os.makedirs(self.model_path)
        fname = self.model_path+"config.cfg"
        with open(fname, 'w') as f:
            f.write('### DeXtrusion configuration file - dont edit ###\n')
            f.write('nb_events_category = '+str(self.ncat)+'\n')
            f.write('events_category_names = '+str(self.catnames)+'\n')
            f.write('window_half_size = '+str(self.half_size)+'\n')
            f.write('nb_temporal_frames = '+str(self.nframes)+'\n')
            f.write('cell_diameter = '+str(self.cell_diameter)+'\n')
            f.write('extrusion_duration = '+str(self.extrusion_duration)+'\n')
            f.write('neuralnetwork_nb_filters = '+str(self.nfilters)+'\n')
            f.write('neuralnetwork_batch_size = '+str(self.batch_size)+'\n')

    def add_configuration_info(self, nepochs=50, naug=1, add_nothing_windows=0, path=""):
        """ Complete the configuration file (file of used parameters) with parameters used for network training 
        :param nepochs: number of iteration to do for training
        :param naug: number of augmentation done to the training data (1 for no augmentation)
        :param add_nothing_windows: add false positive ROIs (saved in file filename_nothing.zip) to force training on this data as control
        :param path: path to the training data
        """
        params = "neuralnetwork_nb_epochs = "+str(nepochs)+"\n"
        params = params + "datatraining_augmentation = "+str(naug)+"\n"
        params = params + "datatraining_addnothingwindows = "+str(add_nothing_windows)+"\n"
        params = params + "datatraining_path = "+path+"\n"
        fname = self.model_path+"config.cfg"
        with open(fname, 'a') as f:
                f.write(params)

    def read_configuration(self, modelpath):
        """ Read configuration file that is automatically saved in the model folder 'modelpath' to set the parameters """
        configfile = open(os.path.join(modelpath,'config.cfg'), 'r')
        lines = configfile.readlines()
        for line in lines:
            vals = line.split("=")
            if len(vals) >= 1:
                if vals[0].strip() == "cell_diameter":
                    self.cell_diameter = float(vals[1])
                if vals[0].strip() == "extrusion_duration":
                    self.extrusion_duration = float(vals[1])
                if vals[0].strip() == "nb_events_category":
                    self.ncat = int(vals[1])
                if vals[0].strip() == "events_category_names":
                    #print(vals[1])
                    self.read_catnames(vals[1])
                    #print(self.catnames)
                if vals[0].strip() == "window_half_size":
                    self.half_size = self.read_tuple(vals[1])
                if vals[0].strip() == "nb_temporal_frames":
                    self.nframes = self.read_tuple(vals[1])
                if vals[0].strip() == "neuralnetwork_nb_filters":
                    self.nfilters = int(vals[1])
                if vals[0].strip() == "neuralnetwork_batch_size":
                    self.batch_size = int(vals[1])
        if self.catnames is None:
            self.catnames = ["", "_cell_death.zip", "_cell_sop.zip"]
        
        #self.group_size = group_size  ## batch of windows that will be runned simultanously in the network (compromise between running speed and GPU utilization)
        
        ## Bounds: size of area to fill with probability of the window calculated (not all the window, around the center)
        self.bounds = ( floor(self.nframes[0]*0.2), floor(self.nframes[1]*0.2)+1, floor(self.half_size[0]*0.3), floor(self.half_size[1]*0.3) )
    
    
    def read_tuple(self, instr):
        return tuple(map(int, re.findall(r'[0-9]+', instr)))

    def read_catnames(self, instr):
        """ Read the event names from the line in the configuration file """
        self.catnames = []
        events = re.split(',|\[|\]|\n', instr)
        for ind in range(1,len(events)-2):
            self.catnames.append(events[ind].strip().replace("'",""))


    ################################################################################################ 
    ###### Neural network (model) calls
    
    def initialize_model(self, nframes=(5,5), half_size=(22,22), ncat=4, nb_filters=8, batch_size=30, model_path="./model"):
        shape = (np.sum(nframes),) + (half_size[0]*2+1, half_size[1]*2+1)
        self.model.create_model(shape, self.ncat, self.nfilters)

    def train_model(self, train_generator, validation_generator, epochs=50, save=True, plot=True):
        """ Train the neural network with the current parameters 
            The trained network will be save in the model_path directory
            :param train_generator: generator that creates the windows used for training from the ROI files
            :param validation_generator: creates the validation windows to check the training evolution
            :param epochs: number of training iteration to do
            ;param save: save the trained network at the end
            :param plot: show the training curves during training
        """
        if self.verbose:
            start_time = time.time()
        self.model.train(train_generator, validation_generator, epochs=epochs, plot=plot)
        if self.verbose:
            print("----- Training in %s minutes ----" % ((time.time()-start_time)/60) )
        if save:
            self.model.save(self.model_path)
            self.write_configuration()
    
    def evaluate_model_prediction(self, test_generator):
        self.model.evaluate_prediction(test_generator)

    def save_model(self):
        self.model.save(self.model_path)

    def load_model(self, modelpath):
        """ Load a model and its parameters
        """
        if self.model is None:
            self.model = Network(self.verbose)
        self.model_path = modelpath
        self.model.reset(modelpath)
        self.read_configuration(modelpath)
    
    def look_at_latent_features(self, test):
        """ Gives access to the layer before the GRU, containint the vector of latent features for each image of the input sequences
        """
        td = self.model.get_timedis_layer()
        if self.verbose:
            print("Latent features shape:")
            print(td.summary())
        print(test.data_path)
        self.model.predict_convolution(test)

    ############################################################################################
    ###### Movie manipulation

    def resize_image(self, img, ratioxy=1, ratioz=1):
        return zoom(img, (ratioz, ratioxy, ratioxy))
    
    def check_image(self, img):
        """ Check the format of the movie
        """
        if len(img.shape) <= 2:
            print("Error, input is not a movie !")
            return -1
        if len(img.shape) > 3:
            print("Error, input has too many dimensions \n Should be a gray scale 3D image, check that it is not RGB")
            return -1
        return 1

    def to_init_shape(self, img):
        """ Resize back to the original movie size
        """
        ratioxy = self.init_shape[1]/img.shape[1]
        ratioz = self.init_shape[0]/img.shape[0]
        return self.resize_image(img, ratioxy=ratioxy, ratioz=ratioz)

    ###########################################################################################
    ###### Detect events from movie: calculate probability map
        
    def divide(self, res, nres):
        """ Divide the proba results by number of windows that filled it """
        nres[nres==0] = 1
        try:
            if self.probamap is not None:
                del self.probamap
                gc.collect()
                self.probamap = None
        except:
            self.probamap = None
        self.probamap = zeros((3,)+self.scaled_shape, dtype="uint8")
        for c in range(self.ncat-1):
            self.probamap[c,:,:,:] = uint8(res[c,(self.nframes[0]):(res.shape[1]-self.nframes[1]),:,:]/nres[(self.nframes[0]):(res.shape[1]-self.nframes[1]),:,:]*255)

    def do_one_group(self, img, pred_channels, npred, group):
        """ Categorize one batch of windows (one group), put the results in pred_channels """
        if self.verbose:
            print("     - Do one group: "+str(group))
        
        # Prepare the batch of windows to run through the network
        nind = min(self.group_size, self.nw[0]*self.nw[1]*self.nw[2]-group)
        img_group = [ self.fill_group(img, group+cur_ind) for cur_ind in range(0,nind) ]
        

        # Do the prediction with the network
        img_group = np.array(img_group)/255 # normalize   
        self.model.reset(self.model_path)
        pred = self.model.predict_batch(img_group)

        # Fill the proba results in the correct place on the pred_channels image
        for c in range(self.ncat-1):
            do = list(map(functools.partial(self.place_element, pred_channels[c], npred, pred[:,1+c], group, c), range(self.group_size)))

    def fill_group(self, img, gind):
        """ Get the window from the original image and place it in the batch of windows to be analysed """
        cur = self.get_index(gind) 
        curimg = img[(cur[0]-self.nframes[0]):(cur[0]+self.nframes[1]), (cur[1]-self.half_size[0]):(cur[1]+self.half_size[0]+1),(cur[2]-self.half_size[1]):(cur[2]+self.half_size[1]+1)]
        
        ## Scale intensities in the window to run through network
        d = npmax(curimg) - npmin(curimg) 
        if (d>0):
            curimg = (curimg - npmin(curimg))/d
        else:
            curimg = (curimg-npmin(curimg))
        return expand_dims( uint8(curimg*255), axis=3)
    
    def get_index(self, gind):
        index = (self.nframes[0] + self.shiftz + int((gind/self.nw[3])%self.nw[0])*self.dzxy[0], self.half_size[0]+ self.shiftxy + int((gind/self.nw[2])%self.nw[1])*self.dzxy[1], self.half_size[1]+ self.shiftxy + int(gind%self.nw[2])*self.dzxy[1])
        return index

    def update_nwins(self, imgshape):
        nx = int(ceil((imgshape[2]-self.half_size[1]*2-1-self.shiftxy)/(self.dzxy[1]*1.0)))
        ny = int(ceil((imgshape[1]-self.half_size[0]*2-1-self.shiftxy)/(self.dzxy[1]*1.0)))
        nz = int(ceil((imgshape[0]-np.sum(self.nframes)-self.shiftz)/(self.dzxy[0]*1.0)))
        self.nw = (nz, ny, nx, ny*nx, nz*ny*nx)

    def place_element(self, pred_c, npred, pred, group, chan, ind):
        """ Place the results of the window in the correct place of the results image """
        # current index
        gind = group+ind
        
        # out of range
        if gind >= self.nw[4]:
            return 0
        
        index = self.get_index(gind)

        pred_c[(index[0]-self.bounds[0]):(index[0]+self.bounds[1]),
                (index[1]-self.bounds[2]):(index[1]+self.bounds[2]+1),
                (index[2]-self.bounds[3]):(index[2]+self.bounds[3]+1)] += pred[ind]
        if chan==0:
            npred[(index[0]-self.bounds[0]):(index[0]+self.bounds[1]),
                (index[1]-self.bounds[2]):(index[1]+self.bounds[2]+1),
                (index[2]-self.bounds[3]):(index[2]+self.bounds[3]+1)] += 1
        return 0
    
    def detect_events_onmovie(self, moviepath, models=[], cell_diameter=25, extrusion_duration=4.5, dxy=25, dz=2, group_size=150000, outfolder=None):
        """ Detect events from the movie by using sliding windows and trained network
        :param moviepath: path to the movie to do the detection on
        :param models: list of the names of one (or more) networks to use for the dection
        :param cell_diameter: typical size (in pixels) of a cell, used to resize if necessary the movie
        :param extrusion_duration: typical temporal duration of an extrusion (number of frames on which it is visible), to resize if necessary
        :param dxy: spatial step of sliding window
        :param dz: temporal step of sliding window
        :param group_size: computational parameter: number of windows runned at the same time for detection. Depends on the computer capacity (more = faster)
        :param outfolder: where to save the results. By default will create a folder called "results" in the same folder as the input movie
        """
        if self.verbose:
            start_time = time.time()
        img = imread(moviepath)
        imdir = os.path.dirname(moviepath)
        output_name = os.path.basename(moviepath)
        output_name = os.path.splitext(output_name)[0]
        if outfolder is None:
            self.outpath = os.path.join(imdir, "results")
        else:
            self.outpath = outfolder
        if not os.path.exists(self.outpath):
            os.makedirs(self.outpath)
        self.outname = os.path.join(self.outpath, output_name)
        self.detect_events(img, models, cell_diameter, extrusion_duration, dxy, dz, group_size=group_size)

        if self.verbose:
            print("---- Detection finished, took %s minutes ---" % ((time.time() - start_time)/60))
        del img
        gc.collect()

    def detect_events(self, img, model_paths=[], cell_diameter=25, extrusion_duration=4.5, dxy=25, dz=2, group_size=150000):
        """ Predict events on an image (already loaded in img parameter, return labelled image """
       
        ### load model parameters and check image
        if len(model_paths) <= 0:
            print("Choose model(s) to use (should be a list) !")
            return
        self.load_model(model_paths[0])
        
        if self.check_image(img) < 0:
            return
        self.group_size = group_size

        ### resize if necessary to have same scale as training
        self.init_shape = img.shape
        if self.verbose:
            print("  Initial image shape: "+str(img.shape))
        if abs(cell_diameter-self.cell_diameter) > (self.cell_diameter*0.3):
            ratio = self.cell_diameter/cell_diameter
            img = self.resize_image(img, ratioxy=ratio, ratioz=1)

        ### resize temporally if necessary
        if abs(extrusion_duration-self.extrusion_duration) > (self.extrusion_duration*0.3):
            ratio = self.extrusion_duration/extrusion_duration
            img = self.resize_image(img, ratioxy=1, ratioz=ratio)
        self.scaled_shape = img.shape
        
        ### add frames before and after to sreen all the times with the windows
        if self.verbose:
            print("  Rescaled image shape: "+str(img.shape))
        first = img[0,]
        img = np.concatenate( (np.repeat(first[np.newaxis,], self.nframes[0], axis=0), img), axis=0)
        last = img[len(img)-1,]
        img = np.concatenate( (img, np.repeat(last[np.newaxis,], self.nframes[1], axis=0)), axis=0)
        if self.verbose:
            print("  Extended image shape: "+str(img.shape))

        ### initialize parameters
        self.dzxy = (dz, dxy)
        pred_channels = zeros((self.ncat-1,)+img.shape, dtype="float16") 
        npred = zeros(img.shape, dtype="uint8")
        
        ## run for each model in the models list
        nmod = len(model_paths)
        for cmod in range(nmod):
            self.shiftz = floor(cmod*dxy/nmod)
            self.shiftxy = floor(cmod*dz/nmod)
            self.load_model(model_paths[cmod])
            self.update_nwins(img.shape)
            if self.verbose:
                print("   Doing windows with shift: "+str(self.shiftxy)+" "+str(self.shiftz)+" and model "+model_paths[cmod])
                print("   Number of windows to categorize: "+str(self.nw[4]))
            do = list(map(functools.partial(self.do_one_group, img, pred_channels, npred), range(0,self.nw[4],self.group_size) ))
        
        ## divide result map by the number of predictions added in each pixel
        self.divide(pred_channels, npred)
        del pred_channels
        del npred
  
    ##################################################################################################
    ####### Postprocess results: probability maps, Fiji ROIs
    
    def rescale_position(self, z, y, x):
        ratioxy = self.init_shape[1]/self.probamap.shape[2]
        ratioz = self.init_shape[0]/self.probamap.shape[1]
        z = floor(z*ratioz)
        y = floor(y*ratioxy)
        x = floor(x*ratioxy)
        return z, y, x

    def resize_probamap(self, cat=1):
        ''' resize to initial size proba map '''
        interp_img = zoom(self.probamap[cat-1], (self.dzxy[0], self.dzxy[1], self.dzxy[1]), grid_mode=False)
        diffz = interp_img.shape[0]-self.init_shape[0]
        res = zeros(self.init_shape, dtype="uint8")
        res[:, self.half_size[0]:(self.half_size[0]+interp_img.shape[1]), self.half_size[1]:(self.half_size[1]+interp_img.shape[2])] = interp_img[0:(interp_img.shape[0]-diffz),:,:]
        return res

    def write_rawproba_maps(self, endname="_rawproba.tif", astime=True):
        """ Save the raw probability maps of all the different event category in .tif movies (not rescaled) """
        for cat in range(1, self.ncat):
            self.write_rawproba_map(cat, endname, astime)

    def write_rawproba_map(self, cat=1, endname="_rawproba.tif", astime=True):
        ''' Draw raw probability map of event category 'cat' in a .tif movie (raw values) '''
        res = self.probamap[cat-1]
        catname = self.catnames[cat]
        catname = os.path.splitext(catname)[0]
        outfile = self.outname+catname+endname
        if astime:
            imwrite(outfile, res, imagej=True, metadata={'axes':'TYX'})
        else:
            imwrite(outfile, res, imagej=True, metadata={'axes':'ZYX'})
        if self.verbose:
            print("Raw probability map "+outfile+" saved")
    
    def write_probamaps(self, imgpath=None, cat=1, astime=True):
        """ Write probability map(s) scaled to original movie size """
        if cat is None:
            cats = range(1,self.ncat)
        else:
            cats = [cat]
        
        for icat in cats:
            if imgpath is None:
                img = self.probamap[icat-1]
            else:
                ### get the movie file directory, name and size
                img = self.get_rawproba_frompath(imgpath, cat=icat)
            self.rescale_and_write( img, cat=icat, endname="_proba.tif", imgtype="Probability map", astime=astime)


    def write_cleanedprobamap(self, cat=1, volume_threshold=800, proba_threshold=180, disxy=10, distime=4, imgpath=None, endname="_proba.tif", astime=True):
        ''' From raw probability map, draw the cleaned and rescaled probability map of event category 'cat' in a .tif movie '''

        res = self.clean_probamap(imgpath, cat=cat, threshold=125, mindxy=disxy, mindt=distime, volume_threshold=volume_threshold, proba_threshold=proba_threshold)
        self.rescale_and_write( res, cat=cat, endname=endname, imgtype="Cleaned probability map", astime=astime )

    def rescale_and_write(self, img, cat=1, endname=".tif", imgtype="Movie", astime=True ):
        """ Rescale if necessary the image to initial image shape, and save it as .tif file """
        if self.init_shape != img.shape:
            if self.verbose:
                print("Rescaling from "+str(img.shape)+" to "+str(self.init_shape))
            img = self.to_init_shape(img)

        catname = self.catnames[cat]
        catname = os.path.splitext(catname)[0]
        outfile = self.outname+catname+endname
        if astime:
            imwrite(outfile, img, imagej=True, metadata={'axes':'TYX'})
        else:
            imwrite(outfile, img, imagej=True, metadata={'axes':'ZYX'})
        if self.verbose:
            print(imgtype+" "+outfile+" saved")
   
    def get_rawproba_frompath( self, imgpath, cat=1):
        ### get the movie file directory, name and size
        im = imread(imgpath)
        self.init_shape = im.shape
        im = None
        self.outpath = os.path.join( os.path.dirname(imgpath), "results")
        output_name = os.path.basename(imgpath)
        output_name = os.path.splitext(output_name)[0]
        catname = self.catnames[cat]
        catname = os.path.splitext(catname)[0]
        probapath = os.path.join(self.outpath, output_name+catname+"_rawproba.tif")
        img = imread(probapath)
        self.outname = os.path.join(self.outpath, output_name)
        return img
    
    def clean_probamap(self, imgpath, cat=1, threshold=125, mindxy=5, mindt=3, volume_threshold=800, proba_threshold=180):
        """ Transform input image to event, looking at object shape """
        if imgpath is None:
            img = self.probamap[cat-1]
        else:
            img = self.get_rawproba_frompath( imgpath, cat=cat)

        bimg, labels, llabels, vols, vals = self.rawproba_to_volumes( img, threshold=125, mindxy=mindxy, mindt=mindt )
        cleanproba = np.zeros(img.shape, dtype="uint8")
        for lab, vol, val in zip(llabels, vols, vals):
            if not isnan(vol):
                if (vol>volume_threshold) and (val>proba_threshold):
                    cleanproba[labels==lab] = floor(val)
        return cleanproba


    def get_rois(self, cat=None, volume_threshold=1000, proba_threshold=180, disxy=10, dist=4):
        """ From probamap, get ROIs from centroids of positive (proba>=proba_threshold) volumes 
        :param cat: number of the event (in the list of events) to process. If None, will process all the events.
        :param volume_threshold: threshold of the positive detection volume to keep it and transform it to ROI
        :param proba_threshold: threshold of probability of the event detection to consider the pixel as positive
        :param disxy: threshold spatial distance to consider two events to be the same
        :param dist: temporal distance to consider two events to be the same
        """
        
        if cat is None:
            cats = range(1,self.ncat)
        else:
            cats = [cat]
        
        for icat in cats:
            binimg, labels, llabels, vols, vals = self.rawproba_to_volumes(self.probamap[icat-1], 125, mindxy=disxy, mindt=dist)
            coords = measurements.center_of_mass(binimg, labels=labels, index=llabels)
            del binimg
            gc.collect()
            rois = []
            for (pt, vol, val) in zip(coords, vols, vals):
                if not isnan(pt[0]) and (vol>volume_threshold) and (val>proba_threshold):
                    roiz, roiy, roix = self.rescale_position( pt[0], pt[1], pt[2] )
                    rois.append(ru.create_roi((roiz, roiy, roix), cat=icat))
            
            outfile = self.outname+self.catnames[icat]
            ru.write_rois(outfile, rois, verbose=self.verbose)

    def rawproba_to_volumes(self, img, threshold, mindxy=5, mindt=3, to_image=False):
        """ Transform raw probamap to separated event volumes (local maxima and watershed) """
        binimg = np.copy(img)
        binimg[img>threshold]=1
        binimg[img<=threshold]=0
        # local center of the contigous positive volume
        binimg = minimum_filter(binimg, size=(0, 2, 2))
        distance = distance_transform_edt(binimg)
        maxs = local_maxima(distance, connectivity=16)
        maxs = maximum_filter(maxs, size=(mindt, mindxy, mindxy))
    
        maxlabels, num_labels = label(maxs>0)
        labels = watershed(-distance, maxlabels, mask=binimg)
        # measure center of mass position and size and intensity of each positive volume
        num_labels = np.max(labels)
        listlabels = np.arange(1,num_labels+1)
        vols = ndsum(binimg, labels, listlabels)
        vals = ndsum(img, labels, listlabels)
        vals = vals/vols
        del distance
        del maxs
    
        return binimg, labels, listlabels, vols, vals

    
    def get_rois_fromrawproba_path(self, imagepath=None, cat=1, volume_threshold=800, proba_threshold=180, disxy=10, dist=4, outfile=None):
        """ 
           Clean the raw probamap, get rois and rescale to original image size 
        """
        if imgpath is None:
            img = self.probamap[cat-1]
        else:
            img = self.get_rawproba_frompath( imgpath, cat=cat)

        bimg, labels, llabels, vols, vals = self.rawproba_to_volumes( img, threshold=125, mindxy=mindxy, mindt=mindt )
        coords = measurements.center_of_mass(bimg, labels=labels, index=llabels)
        rois = []
        for (pt, vol, val) in zip(coords, vols, vals):
            if not isnan(pt[0]) and (vol>volume_threshold) and (val>proba_threshold):
                rois.append(ru.create_roi((floor(pt[0]), floor(pt[1]), floor(pt[2])), cat=cat))
            
        if outfile is None:
            outfile = self.outname+self.catnames[cat]
        ru.write_rois(outfile, rois, verbose=self.verbose)

    def compare_rois(self, cat, gtroisfile, resroisfile=None, distance_xy=15, distance_t=4):
        """ Compare two ROIs files and measure accuracy metrics (precision, recall)
        :param cat: number of the event (in the event names list) to compare the ROIs
        :param gtroisfile: manual (or other source) detections ROIs file
        :param resroisfile: dextrusion output (or other source) detection ROIs file. If None, look for the file in the results folder, where dextrusion save the results
        :param distance_xy: spatial distance (in pixels) to consider two ROIs as the same
        :param distance_t: temporal distance (number of frames) to consider two ROIs as the same
        """
        if resroisfile is None:
            resroisfile = self.outname+self.catnames[cat]
        return ru.compare_rois(resroisfile, gtroisfile, distance_xy=distance_xy, distance_t=distance_t)
    
    def write_falsepositives(self, cat, resfile, gtfile, distance_xy=15, distance_t=4, outputfile=None):
        """ Find all false positives detection and save them in a Fiji file
        """
        if outputfile is None:
            self.outpath = os.path.dirname(resfile)
            output_name = os.path.basename(resfile)
            output_name = os.path.splitext(output_name)[0]
            outputfile = os.path.join(self.outpath, output_name+"_FP.zip")
        ru.write_falsepositives(resfile, gtfile, outputfile=outputfile, distance_xy=distance_xy, distance_t=distance_t)

    def write_falsenegatives(self, cat, resfile, gtfile, distance_xy=15, distance_t=4, outputfile=None):
        """ Find all false negatives detection and save them in a Fiji file
        """
        if outputfile is None:
            self.outpath = os.path.dirname(resfile)
            output_name = os.path.basename(resfile)
            output_name = os.path.splitext(output_name)[0]
            outputfile = os.path.join(self.outpath, output_name+"_FN.zip")
        ru.write_falsenegatives(resfile, gtfile, outputfile=outputfile, distance_xy=distance_xy, distance_t=distance_t)







from tensorflow.keras.utils import Sequence
from tensorflow.keras.utils import to_categorical
import os, math, random
import tifffile
from glob import glob
import numpy as np
from scipy.ndimage.filters import gaussian_filter
from skimage import exposure
from dextrusion.RoiUtils import read_rois

"""
BSD 3-Clause License

Copyright (c) 2022, Gaëlle  LETORT and Alexis Villars
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

class MovieGeneratorFromROI(Sequence):
    """ Generate the input sequences (small and short windows) around ROIs position

    Generate the given windows and save them in a temporary folder, from which it is read during the training of the network. It allows not to load all the data before training, but to read it only when necessary.
    The temporary folder will be deleted at the end ofr training.
    """

    def __init__(self,
                 data_path,
                 batch_size=1,
                 frame_shape=(5,5),
                 win_halfsize = (25,25),
                 val_ratio=0.2,
                 balance=True,
                 ncat=3,
                 cat_names = None,
                 naug=1,
                 augment_withnoise = True,
                 add_nothing_windows = 10,
                 tmp_directory_path = None,
                 verbose = True,
                 **kwargs):

        self.data_path= data_path
        if data_path is not None:
            if not os.path.isdir(data_path):
                raise Exception("Folder "+data_path+" not found")
            self.dir_list = glob(os.path.join(data_path, '*.tif'))
       
        if tmp_directory_path is None:
            randid = random.randint(1,1000000)
            self.tempdir = data_path+"/tmp_data_from_roi_"+str(randid)+"/"
        else:
            self.tempdir = tmp_directory_path
        self.frame_shape = frame_shape
        self.half_size = win_halfsize
        self.shape = (np.sum(frame_shape),)+(2*win_halfsize[0]+1, 2*win_halfsize[1]+1)
        self.batch_size = batch_size
        self.val_ratio = val_ratio
        self.ncat = ncat
        self.augment = naug
        self.augment_noise = augment_withnoise
        self.nothing = 1.2  ## slight shift, nothing more present
        self.add_nothing = add_nothing_windows ## Read _nothing.zip ROI file to reinforce no event windows (FP otherwise) and augment it XX times
        self.balance = True
        self.trainindex = []
        self.valindex = []
        self.verbose = verbose
        if cat_names is None:
            self.catnames = ["", "_cell_death.zip", "_cell_sop.zip", "_cell_division.zip"]
        else:
            self.catnames = cat_names
        _validation = kwargs.get("_validation", None)
        if _validation is None:
            self.initialize()
        else:
            self.catlist = kwargs.get("_cat", None)
            self.filelist = kwargs.get("_files", None)
            self.donoise = kwargs.get("_donoise", None)
            self.trainindex = np.arange(len(self.filelist))
            
    def initialize(self):
        self.catlist = []
        self.filelist = []
        self.donoise = []
        self.create_lists()
        self.split_val()
    
    def contain_roi(self, roi, z, y, x):
        """ check if window of center (z,y,x) contains the roi point """
        if roi[0] < z-self.frame_shape[0] or roi[0] > (z+self.frame_shape[1]):
            return False
        if roi[1] < (y-self.half_size[0]) or roi[1] > (y+self.half_size[0]+1):
            return False
        return roi[2] > (x-self.half_size[1]) and roi[2] < (x+self.half_size[1]+1)
    
    def not_contain_rois(self,rois, z, y, x):
        """ Check that window centered in (z,y,x) don't contain a roi in the list """
        for roi in rois:
            if self.contain_roi(roi, z, y, x):
                return False
        return True
    
    def split_val(self):
        """ Separate train and validation sets """
        npts = len(self.catlist)
        indexes = np.arange(npts)
        np.random.shuffle(indexes)

        nbval = int(self.val_ratio * npts)
        nbtrain = npts - nbval
        val = np.random.permutation(indexes)[:nbval]
        indexes = np.array([i for i in indexes if i not in val])
        self.valindex = val
        self.trainindex = indexes
    
    def create_lists(self):
        """ Load train folder and list windows """
        if self.verbose:
            print("Found files: ")
        
        if not os.path.exists(self.tempdir):
            os.makedirs(self.tempdir)

        num = 0
        for imgname in self.dir_list:
            if self.verbose:
                print(imgname)
            img = tifffile.imread(imgname)
            nz, ny, nx = img.shape
            purename = os.path.basename(imgname)
            basename = imgname[:len(imgname)-4]
            
            ## get number of rois in each category
            crois = []
            nrois = np.zeros(self.ncat)
            for cat in range(1,self.ncat):
                roifilename = basename+self.catnames[cat]
                if os.path.isfile(roifilename):
                    rois = read_rois(roifilename)
                    crois = crois + rois
                    nrois[cat] = len(rois)
     
            if self.verbose:
                print("Nrois "+str(max(nrois)))
              
            ### Get rois of each category and balance/augment them if necessary
            for cat in range(1,self.ncat):
                roifilename = basename+self.catnames[cat]
                if os.path.isfile(roifilename):
                    rois =  read_rois(roifilename)
                    if nrois[cat]<0.75*max(nrois) and self.balance:
                        nrois[cat] = 0.85*max(nrois)
                    num = self.add_and_augment_rois(rois, nrois, cat, num, img, purename)

            ### Get "Nothing" rois
            lim = max(nrois)*self.nothing*self.augment*3000   ## stop after too many trials (can create some imbalance
            it = 0
            ndata = 0
            while ndata < (max(nrois)*self.nothing*self.augment) and (it<lim):
                it += 1
                x =  self.half_size[1] + np.random.randint(nx-2*self.half_size[1])
                y =  self.half_size[0] + np.random.randint(ny-2*self.half_size[0])
                z =  self.frame_shape[0] + np.random.randint(nz-self.frame_shape[1])
                if self.not_contain_rois(crois, z, y, x):
                    rimg = self.get_roi_img(z, y, x, img)
                    if rimg is not None:
                        self.add_point(rimg, 0, num, purename, np.random.uniform(0,1.0)<0.5)
                        ndata += 1
                        num += 1
    
            ### Add ROIs that contain no event, reinforcement to reduce false positive
            if self.add_nothing > 1:
                num = self.add_nothing_rois(basename, num, img, purename)

    def add_nothing_rois(self, basename, num, img, purename):
        roifilename = basename+"_nothing.zip"
        if os.path.isfile(roifilename):
            rois =  read_rois(roifilename)
            for iroi in range(len(rois)):
                z, y, x = self.get_roi_pos(rois, iroi)
                rimg = self.get_roi_img(z, y, x, img)
                if rimg is not None:
                    self.add_point(rimg, 0, num, purename, True)
                    num = num + 1

            ## augment, of double augment times
            if self.augment>1:
                ntarget = len(rois)*self.add_nothing
                ndone = 0
                lim = ntarget*1000   ## stop after too many trials (can create some imbalance
                it = 0
                while (ndone < ntarget) and (it<lim):
                    it = it + 1
                    ind = random.randrange(len(rois))
                    z, y, x = self.get_roi_pos(rois, ind)
                    rimg = self.get_roi_img(z, y, x, img)
                    if rimg is not None:
                        self.add_point(rimg, 0, num, purename, False)
                        num = num + 1
                        ndone = ndone + 1
        return num
        


    def add_and_augment_rois(self, rois, nrois, cat, num, img, purename):
        ncat = 0
        for iroi in range(len(rois)):
            z, y, x = self.get_roi_pos(rois, iroi)
            rimg = self.get_roi_img(z, y, x, img)
            if rimg is not None:
                self.add_point(rimg, cat, num, purename, True)
                num = num + 1
                ncat = ncat + 1

        ## augment if necessary
        lim = nrois[cat]*self.augment*2000   ## stop after too many trials (can create some imbalance
        it = 0
        while (ncat < nrois[cat]*self.augment) and (it<lim):
            it += 1
            ind = random.randrange(len(rois))
            z, y, x = self.get_roi_pos(rois, ind)
            rimg = self.get_roi_img(z, y, x, img)
            if rimg is not None:
                self.add_point(rimg, cat, num, purename, False)
                num = num + 1
                ncat = ncat + 1
        return num


    def get_roi_pos(self, rois, ind):
        """ 
            Get the roi position and add a small random noise around it
        """
        roi = rois[ind]
        ## center position
        xroi = roi[2]
        yroi = roi[1]
        frame = roi[0]
        ## draw randomly around the center point
        y = math.floor(yroi + np.random.uniform(-1.0,1.0)*self.half_size[0]*0.2)
        x = math.floor(xroi + np.random.uniform(-1.0,1.0)*self.half_size[1]*0.2)   
        z = math.floor(frame + np.random.uniform(-1.0,1.0)*2)
        return (z, y, x)


    def add_point(self, im, cat, num, name, orig):
        """
            Add a window in category cat
            Save a tiff image of the window in the temp folder
            And it to the lists
        """
        outname = self.tempdir+"cat"+str(cat)+"_"+name+"_roi"+str(num)+".tif"
        im = np.uint8(self._min_max_scaling(im)*255)
        tifffile.imwrite(outname, im, imagej=True)
        self.filelist.append(outname)
        self.catlist.append(cat)
        if orig:
            self.donoise.append(False)
        else:
            self.donoise.append(True)

    def write_img(self, im, ind, num, folder="conv"):
        """ save image im as a tif file """
        resdir = self.data_path+"/results_"+folder+"/"
        if not os.path.exists(resdir):
            os.makedirs(resdir)
        fname = os.path.basename(self.get_batch_names(ind)[num])
        outname = resdir+fname
        im = np.uint8(self._min_max_scaling(im)*255)
        tifffile.imwrite(outname, im, imagej=True)

    def get_roi_img(self, z, y, x, img):
        """ 
            Get the image window around roi 
            Return None if the roi is in the border (all the window doesn't fit)
        """
        if (x-self.half_size[1])<0 or (x+self.half_size[1]+1)>img.shape[2]:
            return None
        if (y-self.half_size[0])<0 or (y+self.half_size[0]+1)>img.shape[1]:
            return None
        if (z-self.frame_shape[0])<0 or (z+self.frame_shape[1])>img.shape[0]:
            return None
        return img[z-self.frame_shape[0]:z+self.frame_shape[1],y-self.half_size[0]:y+self.half_size[0]+1,x-self.half_size[1]:x+self.half_size[1]+1]

    def get_validation_generator(self):
        return self.__class__(
            data_path = self.data_path,
            batch_size = self.batch_size,
            frame_shape = self.frame_shape,
            win_halfsize = self.half_size,
            val_ratio = self.val_ratio,
            balance = self.balance,
            naug = self.augment,
            augment_noise = self.augment_noise,
            verbose = self.verbose,
            ncat = self.ncat,
            tmp_directory_path = self.tempdir,
            _validation = True,
            _cat = [self.catlist[i] for i in self.valindex],
            _files = [self.filelist[i] for i in self.valindex],
            _donoise = [self.donoise[i] for i in self.donoise]
        )         
                
                
    def __len__(self):
        nfiles = len(self.trainindex)
        return math.floor(nfiles/self.batch_size)
            
    def __getitem__(self, idx):
        img_batch = np.empty((self.batch_size,)+self.shape+(1,))
        cat_batch = [0]*self.batch_size
        
        for batch in range(self.batch_size):
            index = self.trainindex[(batch+idx*self.batch_size)%len(self.trainindex)]
            img = tifffile.imread(self.filelist[index])
            noising = self.donoise[index]
            ## introduce some randomization by flipping the movie
            if np.random.uniform(0,1.0)<0.3:    
                img = np.flip(img,axis=1)
            if np.random.uniform(0,1.0)<0.3:
                img = np.flip(img,axis=2)

            if noising and self.augment>1 and self.augment_noise:
                ## random gama exposure filter
                if np.random.uniform(0,1.0)<0.25:
                    gama = 0.6+0.8*np.random.uniform(0,1.0)
                    img = exposure.adjust_gamma(img, gama)
                
                ## random gama exposure filter on one time-frame only (shadow, brightness)
                if np.random.uniform(0,1.0)<0.3:
                    gama = 1.5*np.random.uniform(0,1.0)
                    frame = np.random.randint(0,np.sum(self.frame_shape))
                    img[frame,] = exposure.adjust_gamma(img[frame,], gama)
                
                
                if np.random.uniform(0,1.0)<0.25:
                    sig = 1.5*np.random.uniform(0,1.0)
                    img = gaussian_filter(img, sigma=sig)
                if np.random.uniform(0,1.0)<0.4:
                    noise = np.random.normal(0,1,img.shape)
                    sig = 1.5*np.random.uniform(0,1.0)
                    noise = gaussian_filter(noise, sigma=sig)
                    img = img + noise

                ## put a random square to 0
                if np.random.uniform(0,1.0)<0.15:
                    sqsize = np.random.randint(1,12)
                    mask = np.full((self.half_size[0]*2+1,self.half_size[1]*2+1), 1)
                    indx = np.random.randint(0,self.half_size[0]*2-sqsize)
                    indy = np.random.randint(0,self.half_size[1]*2-sqsize)
                    mask[indx:(indx+sqsize),indy:(indy+sqsize)] = 0
                    mask = np.repeat(mask[np.newaxis,:,:], np.sum(self.frame_shape), axis=0)
                    img = img * mask
                ## put a random square to 1
                if np.random.uniform(0,1.0)<0.15:
                    sqsize = np.random.randint(1,12)
                    mask = np.full((self.half_size[0]*2+1,self.half_size[1]*2+1), 0)
                    indx = np.random.randint(0,self.half_size[0]*2-sqsize)
                    indy = np.random.randint(0,self.half_size[1]*2-sqsize)
                    mask[indx:(indx+sqsize),indy:(indy+sqsize)] = 1
                    mask = np.repeat(mask[np.newaxis,:,:], np.sum(self.frame_shape), axis=0)
                    img = img + mask


            img_batch[batch,:,:,:,0] = img # np.transpose(src, axes=[2,1,0])
            cat_batch[batch] = self.catlist[index]
        
        cat_batch = to_categorical(cat_batch, self.ncat)
        return self._min_max_scaling(img_batch), cat_batch
    
    def get_batch_names(self, idx):
        fnames = []
        for batch in range(self.batch_size):
            index = self.trainindex[(batch+idx*self.batch_size)%len(self.trainindex)]
            fnames.append(self.filelist[index])
        
        return fnames
   
    def clean_tempdir(self):
        """ Clean the temporary folder """
        import shutil
        shutil.rmtree(self.tempdir)
        
    def on_epoch_end(self):
        np.random.shuffle(self.trainindex)
   
    def _min_max_scaling(self, data):
        n = data - np.min(data)
        d = np.max(data) - np.min(data) 
        if (d>0):
            return n/d
        return n
   




"""
        DeXtrusion: detect events from movie script

        Detect extrusions (or another event) on a movie.

    Used already trained neural network(s) to detect extrusions or other events from movie(s).

    DeXtrusion calculates a probability map which reflects the probability of having an extrusion (or other event) for each pixel at each time of the movie. The position of an event will be at the centroid of a high probability volume.

    
"""

import os, sys
import dextrusion.DialogDeXtrusion as dd
import dextrusion.DeXtrusion as dextrusion
from glob import glob


def get_models(modeldir, talkative=True):
    ''' Get all the models in the "modeldir" if it contains several subdirectory, else take only current model '''
    if not os.path.exists( os.path.join(modeldir,"config.cfg") ):
        return glob(modeldir+"/*/", recursive = False)
    else:
        return [modeldir]


def main():
    # default parameters
    talkative = True  ## print info messages
    groupsize = 150000       ## number of windows that are runned through the neural network at the same time - depends on the computer processing capabilities
                         ## higher will run faster, but too high can fail
    imname = os.getcwd()    ## folder where data are saved to initialize the file browser
    modeldir = os.path.join(os.path.abspath(os.path.join(os.getcwd(), os.pardir)), "DeXNets", "notum_ExtSOPDiv")  ## folder where models are saved to initialize the file browser
    dexter = dextrusion.DeXtrusion(verbose=talkative)

    # Choose parameters through graphical interface
    dexter = dextrusion.DeXtrusion(verbose=talkative)
    diag = dd.DialogDeXtrusion()
    diag.dialog_main(modeldir, imname)

    cell_diameter = float(diag.cell_diameter)
    extrusion_duration = float(diag.extrusion_duration)
    shift_xy = int(diag.dxyval)
    shift_t = int(diag.dtval)
    save_proba = bool(diag.saveProba)
    save_proba_one = bool(diag.saveProbaOne)
    save_rois = bool(diag.saveRois)
    cat = int(diag.event_ind)
    catnames = diag.event
    roi_threshold = float(diag.threshold)  ## min average probability in volume to keep ROI
    roi_volume = float(diag.proba_vol)     ## min positive volume to keep ROI
    groupsize = int(diag.group_size)

    modeldir = diag.modeldir
    model = get_models(modeldir, talkative)

    imname = diag.imagepath
    if talkative:
        print("Using network(s) "+str(model))
        print("On movie(s) "+str(imname))

    ## Merge "peaks" if distance < disxy (spatial) & dist (time) 
    disxy = 10
    distime = 4
    # Parameters: astime: save tif image as temporal stack (else slice stack)
    temporal = True

    for image in imname:
        print(image)
        if os.path.exists(image):
            if talkative:
                print( "Detecting events on movie "+str(image) )
                                                
            ## Detect events
            dexter.detect_events_onmovie( image, models=model, 
                                          cell_diameter=cell_diameter, extrusion_duration=extrusion_duration,                                                                           dxy=shift_xy, dz=shift_t,                                                                                                                   group_size=groupsize )
                                                                          
            ## Save probability maps to file, to visualize the results
            if save_proba:
                dexter.write_probamaps(cat=None, astime=temporal)
                                                                                                        
            ## Save cleaned probability map(s) to file, keeping only events above the size and probability thresholds (for visualisation) 
            if save_proba_one:
                dexter.write_cleanedprobamap(cat=cat, volume_threshold=roi_volume, proba_threshold=roi_threshold, disxy=disxy, distime=distime, astime=temporal)
                                                                                                                                                        ## Look for centroid of "positive" volumes in the probability maps and convert it to a point event 
            ## Keep only ROIs that have a big enough volume ('volume_threshold') and with high enough probability ('proba_threshold') 
            # - put 0 and 0 to keep all detections
            if save_rois:
                dexter.get_rois(cat=cat, volume_threshold=roi_volume, proba_threshold=roi_threshold, disxy=disxy, dist=distime)

    return 1 

if __name__ == '__main__':
        sys.exit(main()) 



import tkinter as tk
from tkinter import filedialog, ttk
import os, re
from glob import glob

class DialogDeXtrusion:
 

    def finish(self):
        self.cell_diameter = self.celld.get()
        self.extrusion_duration = self.extd.get()
        self.dxyval = self.dxy.get()
        self.dtval = self.dt.get()
        self.modeldir = self.dirname.get()
        self.imagepath = self.imname.get()
        self.imagepath = self.tolist(self.imagepath)
        self.saveProba = (self.varprob.get() == 1)
        self.saveProbaOne = (self.varprobone.get() == 1)
        self.saveRois = (self.varroi.get() == 1)
        self.proba_vol = self.pvol.get()
        self.threshold = self.pthres.get()
        self.event = self.list.get()
        self.event_ind = self.list.current()
        self.group_size = self.grsize.get()
        self.window.destroy()
    

    def tolist(self, imagepath):
        res = []
        imagepath = imagepath.replace("(","")
        imagepath = imagepath.replace(")","")
        imagepath = imagepath.replace("'","")
        imagepath = imagepath.replace("\"","")
        imagepath = imagepath.replace("[","")
        imagepath = imagepath.replace("]","")
        images = re.split(',', imagepath)
        #print(images)
        for im in images:
            im = im.lstrip(" ")
            res.append(im)
        return res


    def askmodeldirectory(self):
        self.model = filedialog.askdirectory(initialdir=self.model, title="Go inside model directory")
        self.dirname.delete(0, tk.END)
        self.dirname.insert(0,str(self.model))
        self.catlist = self.read_catnames_fromconfiguration(self.model)
    
    def read_catnames_fromconfiguration(self, modelpath):
        """ Read configuration file that is automatically saved in the model folder 'modelpath' to set the parameters """
        configpath = os.path.join(modelpath,'config.cfg')
        if not os.path.exists(configpath):
            path = glob(modelpath+"/*/", recursive=False)
            print(path)
            if len(path) <=0:
                self.list['values'] = ["", "_cell_death.zip"]
                return
            i = 0
            while not os.path.exists(configpath):
                print(path[i])
                configpath = os.path.join(path[i],'config.cfg')
                i = i + 1
                if i > len(path):
                    self.list['values'] =  ["", "_cell_death.zip"]
                    return

        configfile = open(configpath, 'r')
        lines = configfile.readlines()
        for line in lines:
            vals = line.split("=")
            if len(vals) >= 1:
                if vals[0].strip() == "events_category_names":
                    self.list['values'] = self.read_catnames(vals[1])
                    return
        self.list['values'] = ["", "_cell_death.zip"]
    
    def read_catnames(self, instr):
        catnames = []
        events = re.split(',|\[|\]|\n', instr)
        for ind in range(1,len(events)-2):
            catnames.append(events[ind].strip().replace("'",""))
        return catnames

        
        ## Bounds: size of area to fill with probability of the window calculated (not all the window, around the center)
        self.bounds = ( floor(self.nframes[0]*0.2), floor(self.nframes[1]*0.2)+1, floor(self.half_size[0]*0.3), floor(self.half_size[1]*0.3) )
        
    def askmovie(self):
        self.impath = filedialog.askopenfilenames()
        self.imname.delete(0, tk.END)
        self.imname.insert(0,str(self.impath))
    
    def dialog_main(self, model_dir=os.getcwd(), impath=os.getcwd()):  
        self.window = tk.Tk()
        self.window.title("DeXtrusion")
        self.window.grid_columnconfigure(0, weight=1)
        self.window.grid_columnconfigure(1, weight=1)
        self.window.grid_columnconfigure(2, weight=8)
        self.model = model_dir
        self.impath = impath
        self.list = ttk.Combobox(self.window, values = [])
        self.list.set("Choose event to detect:")

        lastrow = 0
        pathw = 50
        numw = 10
        lastrow = lastrow + 1
        Ltrt = tk.Label(self.window, text="Choose network to use, go inside the model folder for that. If you want to run several network, they all should be in that folder.")
        Ltrt.grid(row=lastrow, column=0, columnspan=4)
        lastrow = lastrow + 1
        
        LM = tk.Label(self.window, text="Trained model(s) to use :")
        LM.grid(row=lastrow, column=0)
        mfile = tk.Button(self.window, text='Browse', command=self.askmodeldirectory)
        mfile.grid(row=lastrow, column=1)
        self.dirname = tk.Entry(self.window, width=pathw)
        self.dirname.grid(row=lastrow, column=2, columnspan=3)
        self.dirname.delete(0,tk.END)
        self.dirname.insert(0, str(self.model))
        self.read_catnames_fromconfiguration(self.model)
        
        lastrow = lastrow + 1
        Lempty = tk.Label(self.window, text="")
        Lempty.grid(row=lastrow, column=0)
        
        lastrow = lastrow + 1
        Ltmt = tk.Label(self.window, text="Choose on which movie(s) to process. You can select several movies.")
        Ltmt.grid(row=lastrow, column=0, columnspan=4)
        
        lastrow = lastrow + 1
        LIm = tk.Label(self.window, text="Input movie(s) :")
        LIm.grid(row=lastrow, column=0)
        img = tk.Button(self.window, text='Choose movie(s):', command=self.askmovie)
        img.grid(row=lastrow, column=1)
        self.imname = tk.Entry(self.window, width=pathw)
        self.imname.grid(row=lastrow, column=2, columnspan=3)
        self.imname.delete(0,tk.END)
        self.imname.insert(0, str(self.impath))
        
        lastrow = lastrow + 1
        empty = tk.Label(self.window, text="")
        empty.grid(row=lastrow, column=0)
        
        lastrow = lastrow + 1
        Lpar = tk.Label(self.window, text="Movie resolution:")
        Lpar.grid(row=lastrow, column=0, columnspan=2)
        
        Lpart = tk.Label(self.window, text="Windows shift:")
        Lpart.grid(row=lastrow, column=3, columnspan=2)
        
        
        lastrow = lastrow + 1
        Lc = tk.Label(self.window, text="Typical cell diameter (pixels):")
        self.celld = tk.Entry(self.window, width=numw)
        self.celld.insert(0,"25.0")
        Lc.grid(row=lastrow, column=0)
        self.celld.grid(row=lastrow, column=1)
        
        empty2 = tk.Label(self.window, text="", width=10)
        empty2.grid(row=lastrow, column=2)
        
        Lxy= tk.Label(self.window, text="spatial shift:")
        self.dxy= tk.Entry(self.window, width=numw)
        self.dxy.insert(0,"10")
        Lxy.grid(row=lastrow, column=3)
        self.dxy.grid(row=lastrow, column=4)
        
        lastrow = lastrow + 1
        Ld= tk.Label(self.window, text="Typical extrusion duration (frames):")
        self.extd = tk.Entry(self.window, width=numw)
        self.extd.insert(0,"4.5")
        Ld.grid(row=lastrow, column=0)
        self.extd.grid(row=lastrow, column=1)

        Lt= tk.Label(self.window, text="Temporal shift:")
        self.dt= tk.Entry(self.window, width=numw)
        self.dt.insert(0,"2")
        Lt.grid(row=lastrow, column=3)
        self.dt.grid(row=lastrow, column=4)
        
        lastrow = lastrow + 1
        empty3 = tk.Label(self.window, text="")
        empty3.grid(row=lastrow, column=0)
        
        lastrow = lastrow + 1
        Lpar = tk.Label(self.window, text="Results output:")
        Lpar.grid(row=lastrow, column=0, columnspan=2)
        
        lastrow = lastrow + 1
        Lpar = tk.Label(self.window, text="Event to ROI:")
        Lpar.grid(row=lastrow, column=0)
        
        self.list.grid(row=lastrow, column=1)
        self.list.set('_cell_death.zip')
        
        lastrow = lastrow + 1
        self.varroi = tk.IntVar(value=1)
        croi = tk.Checkbutton(self.window, text='Fiji ROIs of event \t     ',variable=self.varroi, onvalue=1, offvalue=0)
        croi.grid(row=lastrow, column=0)
        
        Lvol = tk.Label(self.window, text="Min probability volume:")
        self.pvol = tk.Entry(self.window, width=numw)
        self.pvol.insert(0,"800")
        Lvol.grid(row=lastrow, column=1)
        self.pvol.grid(row=lastrow, column=2)
        
        Lthres = tk.Label(self.window, text="Min proba threshold (150-255):")
        self.pthres = tk.Entry(self.window, width=numw)
        self.pthres.insert(0,"180")
        Lthres.grid(row=lastrow, column=3)
        self.pthres.grid(row=lastrow, column=4)
        
        lastrow = lastrow + 1
        self.varprob = tk.IntVar(value=1)
        cprob = tk.Checkbutton(self.window, text='Save all probability maps',variable=self.varprob, onvalue=1, offvalue=0)
        cprob.grid(row=lastrow, column=0, columnspan=2)
        
        self.varprobone = tk.IntVar(value=0)
        cprobone = tk.Checkbutton(self.window, text='Save cleaned selected event proba map',variable=self.varprobone, onvalue=1, offvalue=0)
        cprobone.grid(row=lastrow, column=2, columnspan=2)

        lastrow = lastrow + 1
        Lempy = tk.Label(self.window, text="")
        Lempy.grid(row=lastrow, column=0)
        
        lastrow = lastrow + 1
        Ltgst = tk.Label(self.window, text="Computing efficiency, depends on your computer capacity. The higher the faster, but might exceed memory.")
        Ltgst.grid(row=lastrow, column=0, columnspan=2)
        lastrow = lastrow + 1
        
        Lgr= tk.Label(self.window, text="Computing size:")
        Lgr.grid(row=lastrow, column=0)
        self.grsize= tk.Entry(self.window, width=numw)
        self.grsize.insert(0,"150000")
        self.grsize.grid(row=lastrow, column=1)


        lastrow = lastrow + 1
        but = tk.Button(self.window, text = 'OK', command = self.finish)#.grid(column = 1, row = 3)
        but.grid(row=lastrow, column=5)
        self.window.mainloop()



import os, random, math
import numpy as np
import matplotlib.pyplot as plt
from tensorflow.keras.layers import Conv2D, BatchNormalization, MaxPool2D, GlobalMaxPool2D, Input
from tensorflow.keras.layers import TimeDistributed, GRU, Dense, Dropout
from tensorflow.keras.models import Model
from tensorflow.keras.models import save_model, load_model
from tensorflow.keras.callbacks import Callback, ReduceLROnPlateau, ModelCheckpoint 
from tensorflow.keras.optimizers import SGD
from sklearn import metrics
from IPython.display import clear_output
import tensorflow as tf
from tensorflow.keras.backend import clear_session
from tensorflow.keras.backend import function

"""
BSD 3-Clause License

Copyright (c) 2022, Gaëlle  LETORT and Alexis Villars
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""

class Network:
    """ DeXNet: dextrusion neural network
    
    Allow to configure, train or retrain DeXNets. This class is called from DeXtrusion main functions.
    """
    def __init__(self, verbose=True):
        self.init_tf(verbose)
        self.model = None

    def create_model(self, shape=(10,45,45), ncat=4, nb_filters=8):
        self.model = self.action_model(shape, ncat, nb_filters)
        self.model.compile( SGD(0.1), 'categorical_crossentropy', metrics=['acc'] )

    def save(self, model_path):
        if not os.path.exists(model_path):
            os.makedirs(model_path)
        save_model(self.model, model_path)
        
    def init_tf(self, verbose=True):
        if verbose:
            print("Tensorflow with Cuda: "+str(tf.test.is_built_with_cuda()))
            print("Tensorflow version: "+str(tf.__version__))
            print("Num GPUs Available: ", len(tf.config.list_physical_devices('GPU')))
        config = tf.compat.v1.ConfigProto( device_count = {'GPU': 0} )
        sess = tf.compat.v1.Session(config=config)
        tf.compat.v1.keras.backend.set_session(sess)
        
    def reset(self, model_path):
        clear_session()
        self.model = load_model(model_path)

    

    #### Neural network architecture functions
    
    def get_timedis_layer(self):
         input_tensor = Input(self.shape[1:]+(1,), name='input')
         #layer = Model(inputs=self.model.input,
         #                              outputs=self.model.get_layer("dense_1").output)
         layer = self.model.get_layer("time_distributed").output
         return Model(self.model.input, layer)
    
    def get_last_layer(self, x):
        layer = self.model.get_layer("dense").output
        get_layer_output = function(self.model.layer[0].input, layer)
        return get_layer_output([x])
    
    def conv_block(self,input_tensor, nfil, momentum):
        x = Conv2D(nfil, (3,3), padding='same', activation='relu')(input_tensor)
        x = Conv2D(nfil, (3,3), padding='same', activation='relu')(x)
        x = BatchNormalization(momentum=momentum)(x)
        return x
        
    def build_convnet(self,input_tensor, nfilters):
        momentum = .95
        c1 = self.conv_block(input_tensor, nfilters, momentum)
        c1 = MaxPool2D()(c1)
        c2 = self.conv_block(c1, nfilters*2, momentum)
        c2 = MaxPool2D()(c2)
        c3 = self.conv_block(c2, nfilters*4, momentum)
        c3 = MaxPool2D()(c3)
        c4 = self.conv_block(c3, nfilters*8, momentum)
        c4 = GlobalMaxPool2D()(c4)
        return Model(inputs=[input_tensor], outputs=[c4])

    def action_model(self, shape, ncat, nfilters):
        input_tensor = Input(shape[1:]+(1,), name='input')
        convnet = self.build_convnet(input_tensor, nfilters)
        
        inputl = Input(shape+(1,), name='input')
        tconv = TimeDistributed(convnet)(inputl) 
        de = GRU(64)(tconv)   # 64 GRU or LSTM
        #tconv = LSTM(24)(tconv)   # GRU or LSTM
        
        # Decision network
        de = Dropout(.5)(de)
        de = Dense(32, activation='relu')(de)
        de = Dense(16, activation='relu')(de)
        #de = Dropout(.5)(de)
        #de = Dense(8, activation="relu")(tconv)
        output = Dense(ncat, activation='softmax')(de)
        return Model(inputs=[inputl], outputs=[output])
    
    #### Neural network training functions

    def train(self, train_gen, validation_gen, epochs, plot=True):
        if plot:
            plot_losses = TrainingPlot()
            callbacks = [ ReduceLROnPlateau(verbose=1), plot_losses ]
            self.model.fit(train_gen, validation_data=validation_gen, verbose=1, epochs=epochs, callbacks=callbacks)
        else:
            self.model.fit(train_gen, validation_data=validation_gen, verbose=1, epochs=epochs)
    

    #### Prediction/Evaluation functions

    def evaluate_prediction(self, datatest):
        ''' Evaluate prediction score function '''
        for i in range(0, len(datatest)):
            temp_var = datatest.__getitem__(i)
            a = temp_var[0]
            b = self.model.predict(a)
            if i == 0:
                y_pred = np.argmax(b, axis = 1)
                y_true = np.argmax(temp_var[1], axis = 1)
            else:
                y_pred = np.append(y_pred,np.argmax(b, axis = 1))
                y_true = np.append(y_true, np.argmax(temp_var[1], axis = 1))
    
        print("---------------------------")
        print("-- Scores")
        print("Nb test data: "+str(y_true.shape))  
        print("Accuracy: "+str(metrics.accuracy_score(y_true, y_pred)))
        print("Balanced accuracy: "+str(metrics.balanced_accuracy_score(y_true, y_pred)))
        print("Confusion matrix:")
        print(metrics.confusion_matrix(y_true, y_pred))

    def evaluate_batch(self, datatest, i, what='false_death'):
        temp_var = datatest.__getitem__(i)
        a = temp_var[0]
        b = self.model.predict(a)
        y_pred = np.argmax(b, axis = 1)
        y_true = np.argmax(temp_var[1], axis = 1)
        
        if what == 'false_death':
            return (np.intersect1d(np.where(y_true!=1)[0], np.where(y_pred==1)[0]))
        if what == 'false_sop':
            return (np.intersect1d(np.where(y_true!=2)[0], np.where(y_pred==2)[0]))
        if what == 'nothing_death':
            return (np.intersect1d(np.where(y_true==0)[0], np.where(y_pred==1)[0]))
        return None
        
    def predict_batch(self, img_batch):
        return self.model.predict(img_batch)
        
    def predict_convolution(self, datatest):
        ''' Evaluate prediction score function '''
        td = self.get_timedis_layer()
        for i in range(0, len(datatest)):
            temp_var = datatest.__getitem__(i)
            a = temp_var[0]
            b = td.predict(a)
            j = 0
            for im in b:
                datatest.write_img(im, i, j)
                j = j + 1

    

class TrainingPlot(Callback):
    ''' Handle the visualization of the training loss and accuracy at each epochs ''' 

    def on_train_begin(self, logs={}):
        ''' Initialize the logs when training begins'''
        self.losses = []
        self.acc = []
        self.val_losses = []
        self.val_acc = []
        self.logs = []
    
    def on_epoch_end(self, epoch, logs={}):
        ''' Called at the end of each epoch plot the loss and accuracy '''
        # Append the logs, losses and accuracies to the lists
        self.logs.append(logs)
        self.losses.append(logs.get('loss'))
        self.acc.append(logs.get('acc'))
        self.val_losses.append(logs.get('val_loss'))
        self.val_acc.append(logs.get('val_acc'))
        
        # Before plotting ensure at least 2 epochs have passed
        if len(self.losses) > 1:
            # Clear the previous plot
            clear_output(wait=True)
            N = np.arange(0, len(self.losses))
            # print(plt.style.available) to see the available options
            plt.style.use("seaborn")
            
            # Plot train loss, train acc, val loss and val acc against epochs passed
            fig, axs = plt.subplots(1, 2)
            axs[0].plot(N, self.losses, label = "train_loss")
            axs[0].plot(N, self.val_losses, label = "val_loss")
            axs[1].plot(N, self.acc, label = "train_acc")
            axs[1].plot(N, self.val_acc, label = "val_acc")
            
            axs[0].set_title("Loss SGD [Epoch {}]".format(epoch))
            axs[0].set(xlabel="Epoch #", ylabel="loss")
            axs[1].set_title("accuracy [Epoch {}]".format(epoch))
            axs[1].set(xlabel="Epoch #", ylabel="accuracy")
            axs[1].set_ylim([0.3,1])
            plt.legend()
            plt.show()




from setuptools import setup, find_packages

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='dextrusion',
    long_description=long_description,
    long_description_content_type='text/markdown',
    version='0.0.5',
    description='DeXtrusion: automatic detection of cell extrusion in epithelial tissu',
      author='Gaëlle Letort and Alexis Villars',
      url='https://gitlab.pasteur.fr/gletort/dextrusion',
      package_dir={'':'src'},
      packages=find_packages('src'),
    install_requires=[
        "matplotlib",
        "numpy",
        "opencv-python",
        "tifffile>=2022.2.2",
        "roifile",
        "scikit-image",
        "scikit-learn",
        "scipy",
        "tensorflow==2.8", 
        "protobuf==3.19",
        "ipython"
    ],
)





import tkinter as tk
from tkinter import filedialog
import os

class DialogParameters:
    
    def finish(self):
        self.bs = self.a.get()
        self.epochs = self.ep.get()
        self.scaling = self.sc.get()
        self.aug = self.ag.get()
        self.ratio = self.val.get()
        self.window.destroy()
    
    def askmodeldirectory(self):
        self.modelf = filedialog.askdirectory(initialdir=os.getcwd(), title="Go inside model to retrain directory")
    
    def askdatadirectory(self):
        self.path = filedialog.askdirectory(initialdir=os.getcwd(), title="Go inside retraining data directory")

    def dialog_retrain(self):    
        self.window = tk.Tk()
        self.window.title("GUI")
        
        L4 = tk.Label(self.window, text="Model to retrain:")
        L4.grid(row=0, column=0)
        self.modelf = os.getcwd()
        mfile = tk.Button(self.window, text='Browse', command=self.askmodeldirectory)
        mfile.grid(row=0, column=1)
        
        L6 = tk.Label(self.window, text="Retrain dataset path:")
        L6.grid(row=0, column=2)
        self.path = os.getcwd()
        dfile = tk.Button(self.window, text='Browse', command=self.askdatadirectory)
        dfile.grid(row=0, column=3)

        L1 = tk.Label(self.window, text="Batch size:")
        self.a = tk.Entry(self.window)
        self.a.insert(0,"50")
        L1.grid(row=1, column=0)
        self.a.grid(row=1, column=1)

        
        L2 = tk.Label(self.window, text="Nb epochs to retrain:")
        self.ep = tk.Entry(self.window)
        self.ep.insert(0,"10")
        L2.grid(row=1, column=2)
        self.ep.grid(row=1, column=3)
        
        L3 = tk.Label(self.window, text="Images scaling (in µm):")
        self.sc = tk.Entry(self.window)
        self.sc.insert(0,"0.275")
        L3.grid(row=2, column=0)
        self.sc.grid(row=2, column=1)
        
        L5 = tk.Label(self.window, text="Augment data (1 no aug., >1 aug.):")
        self.ag = tk.Entry(self.window)
        self.ag.insert(0,"3")
        L5.grid(row=2, column=2)
        self.ag.grid(row=2, column=3)
        
        L7 = tk.Label(self.window, text="Train/Validation split ratio:")
        self.val = tk.Entry(self.window)
        self.val.insert(0,"0.2")
        L7.grid(row=3, column=0)
        self.val.grid(row=3, column=1)


        
        but = tk.Button(self.window, text = 'OK', command = self.finish)#.grid(column = 1, row = 3)
        but.grid(row=3, column=2)
        self.window.mainloop()
